from pathlib import Path

from pydantic import Field, validator

from .alias import DEFAULT_ALIAS_FILENAME, AliasConfig
from .base_config import BaseYamlConfigModelWithBase, default_if_none
from .config import Config
from .settings import Settings


class GlobalConfig(BaseYamlConfigModelWithBase):
    defaults: Config = Field(
        default_factory=Config,
        description="default values for all configs for the current environment",
    )
    template_dir: Path = Field(
        Path("./templates"),
        description="directory for global templates",
    )
    override_config_path: Path = Field(
        Path("overrides.yaml"),
        description=("default path for per-project override config (. is in '.devcontainer/')"),
    )

    _not_none = validator("*", pre=True, allow_reuse=True)(default_if_none)

    def load_alias_config(self):
        alias_config_path = self.config_path.parent / DEFAULT_ALIAS_FILENAME
        if not alias_config_path.exists():
            AliasConfig().write_yaml(alias_config_path)

        return AliasConfig.parse_file(alias_config_path)

    @classmethod
    def load(
        cls, settings: Settings = Settings(), create_if_not_exist: bool = False
    ) -> "GlobalConfig":
        config_path = settings.global_config_path
        if not config_path.exists():
            if create_if_not_exist:
                global_config = GlobalConfig()
                global_config.write_yaml(config_path, with_descriptions=True)
                global_config.config_path = config_path
                return global_config
            return None
        return GlobalConfig.parse_file(config_path)



import io
from pathlib import Path

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.error import CommentMark
from ruamel.yaml.tokens import CommentToken

yaml = YAML(pure=True)
yaml.default_flow_style = False
yaml.width = 4096
yaml.indent(mapping=4, sequence=4, offset=2)


def yaml_to_str(dict):
    buf = io.StringIO()
    yaml.dump(dict, buf)
    return buf.getvalue()


yaml.dump_str = yaml_to_str


def represent_none(representer, _):
    return representer.represent_scalar("tag:yaml.org,2002:null", "")


def represent_as_str(representer, data):
    return representer.represent_scalar("tag:yaml.org,2002:str", str(data))


yaml.representer.add_representer(type(None), represent_none)
yaml.representer.add_multi_representer(Path, represent_as_str)
yaml.representer.add_multi_representer(str, represent_as_str)


if not hasattr(CommentedMap, "yaml_set_comment_before_key"):

    def yaml_set_comment_before_key(self, key, comment, column=None):
        key_comment = self.ca.items.setdefault(key, [None, [], None, None])
        comment_list = key_comment[1]
        if comment_list is None:
            key_comment[1] = []
            comment_list = key_comment[1]
        if comment:
            comment_start = "# "
            if comment[-1] == "\n":
                comment = comment[:-1]  # strip final newline if there
        else:
            comment_start = "#"
        if column is None:
            if comment_list:
                # if there already are other comments get the column from them
                column = comment_list[-1].start_mark.column
            else:
                column = 0

        comment = f"\n{' ' * (column)}# ".join(comment.splitlines())
        start_mark = CommentMark(column)
        comment_list.append(CommentToken(comment_start + comment + "\n", start_mark, None))
        return self

    CommentedMap.yaml_set_comment_before_key = yaml_set_comment_before_key

if not hasattr(CommentedMap, "yaml_add_newline_after_key"):

    def yaml_add_newline_after_key(self, key, n=1):
        key_comment = self.ca.items.setdefault(key, [None, None, None, None])
        key_comment[2] = CommentToken("\n" * (n + 1), CommentMark(0), None)
        return self

    CommentedMap.yaml_add_newline_after_key = yaml_add_newline_after_key


if not hasattr(CommentedMap, "yaml_add_newline_before_key"):

    def yaml_add_newline_before_key(self, key, n=1):
        key_comment = self.ca.items.setdefault(key, [None, [], None, None])
        key_comment[1].insert(0, CommentToken("\n" * n, CommentMark(0), None))
        return self

    CommentedMap.yaml_add_newline_before_key = yaml_add_newline_before_key



from pathlib import Path
from typing import List, Optional

from pydantic import Field, root_validator, validator

from .base_config import BaseYamlConfigModel, BaseYamlConfigModelWithBase, default_if_none
from .types import MountString
from .util import get_project_root_basename, render_recursive_template


class DevcontainerConfig(BaseYamlConfigModel):
    name: Optional[str] = Field("{{ project_root_basename }}", description="dev container name")

    workspace_folder: Optional[Path] = Field(
        Path("/mnt/workspace"),
        description="path in container where source will be mounted",
    )

    workspace_mount: Optional[MountString] = Field(
        MountString("${localWorkspaceFolder}:{{ devcontainer.workspace_folder }}"),
        description=(
            "same as workspaceMount in devcontainer.json - path for workspace and where\n"
            "to mount it; there are two available formats:\n"
            "   - same as devcontainer.json\n"
            "   - shortened form - '<local-path>:<remote-path>' - this will be translated\n"
            "     to 'src=<local-path>,dst=<remote-path>,type=bind,consistency=cached'"
        ),
    )
    shutdown_action: Optional[str] = Field(
        "none",
        description="same as shutdownAction in devcontainer.json",
    )
    user_env_probe: Optional[str] = Field(
        "loginInteractiveShell",
        description="same as userEnvProbe in devcontainer.json",
    )
    image: Optional[str] = Field(
        "{{ devcontainer.name }}-dev",
        description="devcontainer image to use",
    )
    mounts: Optional[List[MountString]] = Field(
        default_factory=list,
        description=(
            "additional mounts for container in format: `src:dst`, for example this\n"
            "will mount home folder to /mnt/home in the container\n"
            "mounts:\n"
            "   - /home/developer:/mnt/home"
        ),
    )
    container_name: Optional[str] = Field(
        "{{ devcontainer.name }}",
        description=(
            "name of the container - this will be passed as `--name <arg>` in `docker run`"
        ),
    )
    container_hostname: Optional[str] = Field(
        "{{ devcontainer.name }}",
        description=(
            "container hostname - this will be passed as `--hostname <arg>` in `docker run`\n"
            "this option is to make shell display the hostname as specified name instead\n"
            "of randomly generated container hex code"
        ),
    )
    run_args: Optional[List[str]] = Field(
        default_factory=list,
        description=(
            "aditional arguments that will be passed to `docker run` - i.e. adding gpus:\n"
            "run_args:\n"
            "- gpus=all"
        ),
    )

    extensions: Optional[List[str]] = Field(
        default_factory=list,
        description=(
            "default extensions to install - will be directly translated to devcontainer.json\n"
            "extensions"
        ),
    )
    additional_options: Optional[List[str]] = Field(
        default_factory=list,
        description=(
            "list of additional options to that will be appended to devcontainer config\n"
            "for example:\n"
            "additional_options:\n"
            "  - >\n"
            '    "dockerFile": "{{ docker.file }}"\n'
            '  - \'"appPort": "8080"\''
        ),
    )

    _not_none = validator("*", pre=True, allow_reuse=True)(default_if_none)


class DockerConfig(BaseYamlConfigModel):
    file: Optional[Path] = Field(
        None,
        description=(
            "path for base dockerfile to use for building custom image\n"
            "null means that the dockerfile will not be generated\n"
            "if the path is valid, two files will be generated - devcontainer.Dockerfile\n"
            "and build.sh script for building this dockerfile"
        ),
    )

    additional_commands: Optional[List[str]] = Field(
        default_factory=list,
        description=(
            "additional lines to append to dockerfile - this is useful if the main dockerfile\n"
            "does not contain developer tools, for example to add fish and git:\n"
            "\n"
            "additional_commands:\n"
            "- >\n"
            "  RUN apt-get update && apt-get install\n"
            "     fish procps git git-lfs\n"
            "  && rm -rf /var/lib/apt/lists/*\n"
            '- ENV SHELL="/usr/bin/fish"\n'
            "- ENV LANG=C.UTF-8 LANGUAGE=C.UTF-8 LC_ALL=C.UTF-8\n"
            '- SHELL ["fish", "--command"]\n'
            '- ENTRYPOINT ["fish"]'
        ),
    )

    _not_none = validator("*", pre=True, allow_reuse=True)(default_if_none)


class Config(BaseYamlConfigModelWithBase):
    path: Path = Field(
        Path(".devcontainer"),
        description="custom path for devcontainer settings",
    )
    devcontainer: Optional[DevcontainerConfig] = DevcontainerConfig()
    docker: Optional[DockerConfig] = DockerConfig()

    _not_none = validator("*", pre=True, allow_reuse=True)(default_if_none)

    def resolve(self) -> "ResolvedConfig":
        values = self.dict()
        values["project_root_basename"] = get_project_root_basename()

        cfg_copy: Config = self.copy(deep=True)
        cfg_copy.devcontainer.workspace_mount = (
            self.devcontainer.workspace_mount.to_devcontainer_format()
        )
        cfg_copy.devcontainer.mounts = [
            mount.to_devcontainer_format() for mount in cfg_copy.devcontainer.mounts
        ]

        rendered_template = render_recursive_template(cfg_copy.json(), values)
        new_config = Config.parse_raw(rendered_template)
        new_config.config_path = self.config_path
        return ResolvedConfig.parse_obj(new_config.dict(exclude={"config_path": {}}))

    @classmethod
    @property
    def NONE(cls):
        return Config.construct(
            base_config=None,
            path=None,
            devcontainer=DevcontainerConfig.NONE,
            docker=DockerConfig.NONE,
        )


class ResolvedConfig(Config):
    @root_validator
    def validate_nested(cls, values):
        cls._validate_docker_path(values)
        return values

    def _validate_docker_path(values):
        docker: DockerConfig = values.get("docker")
        config_path: Path = values.get("config_path").resolve().parent
        docker_config_path = docker.file
        if docker_config_path is None:
            return

        if not docker_config_path.is_absolute():
            docker_config_path = (config_path / docker_config_path).resolve()

        if not docker_config_path.exists():
            raise ValueError(
                f"invalid value for 'docker.file' - path '{docker_config_path}' " "does not exist"
            )
        docker.file = Path(docker_config_path).read_text()



from pathlib import Path
from typing import List

import typer

from ..global_config import GlobalConfig

app = typer.Typer()


def complete_aliases(ctx: typer.Context, args: List[str], incomplete: str):
    global_config = GlobalConfig.load()
    param_aliases = ctx.params.get("alias") or []
    for alias, path in global_config.load_alias_config().aliases.items():
        if alias.startswith(incomplete) and alias not in param_aliases:
            yield (alias, path)


@app.command()
def add(alias: str = typer.Argument(...), config_path: str = typer.Argument(...)):
    path = Path(config_path)
    alias_config = GlobalConfig.load().load_alias_config()
    alias_config.aliases[alias] = path.resolve().as_posix()
    alias_config.write_yaml()


@app.command()
def remove(
    alias: List[str] = typer.Argument(..., shell_complete=complete_aliases),
):
    alias_config = GlobalConfig.load().load_alias_config()

    for alias_remove in alias:
        if alias_remove in alias_config.aliases:
            alias_config.aliases.pop(alias_remove)
            alias_config.write_yaml()
        else:
            message = (
                f"{typer.style('Error: ', fg=typer.colors.RED)}"
                "Nothing removed, alias "
                f"'{typer.style(alias_remove, fg=typer.colors.BLUE)}' "
                "does not exist"
            )
            typer.echo(message)


@app.command()
def list():
    global_config = GlobalConfig.load()
    typer.echo("Available aliases:")
    aliases_msg = ""
    for alias, path in global_config.load_alias_config().aliases.items():
        file_exists = Path(path).exists()
        file_color = typer.colors.GREEN if file_exists else typer.colors.RED
        aliases_msg += (
            f"  {typer.style(alias, fg=typer.colors.BLUE)}: "
            f"{typer.style(path, fg=file_color)}"
            "\n"
        )
    typer.echo(aliases_msg)



from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, List, Optional, Type, TypeVar, Union

from pydantic import BaseModel, Field, validator
from pydantic.fields import ModelField
from pydantic_yaml import YamlModel

from .util import dict_merge
from .yaml import yaml

if TYPE_CHECKING:
    Model = TypeVar("Model", bound="BaseModel")


def default_if_none(cls, v, field: ModelField):
    if v is None:
        if getattr(field, "default", None) is not None:
            return field.default
        if field.default_factory is not None:
            return field.default_factory()
    else:
        return v


class BaseYamlConfigModel(YamlModel):
    config_path: Path = Field(Path("."), exclude=True)

    def __or__(self, other: "Model") -> "Model":
        return type(self).parse_obj(
            dict_merge(
                self.dict(exclude={"config_path": {}}, exclude_defaults=True),
                other.dict(exclude={"config_path": {}}, exclude_defaults=True),
            )
        )

    @classmethod
    def parse_file(
        cls: Type["Model"],
        path: Union[str, Path],
        **kwargs,
    ) -> "Model":
        obj = super(BaseYamlConfigModel, cls).parse_file(path, **kwargs)
        obj.config_path = Path(path)
        return obj

    def yaml(self, with_descriptions=False, **dict_kwargs):
        if with_descriptions:
            yaml_dict = yaml.load(self.yaml(**dict_kwargs))
            _construct_yaml_dict_with_comments(type(self), yaml_dict)
        else:
            yaml_dict = self.dict(**dict_kwargs)
        return yaml.dump_str(yaml_dict)

    def write_yaml(self, path: Path = None, create_parents: bool = True, **yaml_kwargs):
        if path is None:
            path = self.config_path
        if create_parents:
            path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(self.yaml(**yaml_kwargs))

    @classmethod
    @property
    def NONE(cls: Type["Model"]) -> "Model":
        return cls.construct(**{field: None for field in cls.__fields__})


class BaseYamlConfigModelWithBase(BaseYamlConfigModel):
    base_config: Optional[List[Path]] = Field(default_factory=list)

    @validator("base_config", pre=True)
    def ensure_base_config_type(cls, v):
        if v is None:
            return []
        if isinstance(v, str):
            return [Path(v)]
        if isinstance(v, Iterable):
            return list(v)
        if isinstance(v, Path):
            return [v]
        raise ValueError(f"invalid value for base_config: '{v}'")

    def merge_bases(self):
        base_configs = [
            config_path if config_path.is_absolute() else self.config_path.parent / config_path
            for config_path in self.base_config
        ]
        base_configs = [
            type(self).parse_file(config_path).merge_bases() for config_path in base_configs
        ]
        if not base_configs:
            return self

        bases_merged = reduce(lambda a, b: a | b, base_configs)
        return bases_merged | self


def _construct_yaml_dict_with_comments(cls, d, column=0):
    for k, v in d.items():
        field = cls.__fields__[k]
        if issubclass(field.type_, BaseModel):
            _construct_yaml_dict_with_comments(field.type_, v, column + 4)

        if field.field_info.description:
            d.yaml_set_comment_before_key(k, field.field_info.description, column=column)
            d.yaml_add_newline_before_key(k)



import subprocess
from functools import reduce
from pathlib import Path
from typing import List, Optional

import typer
from cookiecutter.main import cookiecutter

from .. import __version__
from ..config import Config
from ..global_config import GlobalConfig
from . import alias

TEMPLATE_DIR = Path(__file__).parent.parent / "templates"
COOKIECUTTER_CONFIG = TEMPLATE_DIR / "cookiecutter.json"


app = typer.Typer()
app.add_typer(alias.app, name="alias")


@app.command()
def generate(
    templates: Optional[List[str]] = typer.Argument(None),
    build: bool = typer.Option(False, "--build", "-b"),
):
    global_config = GlobalConfig.load(create_if_not_exist=True)
    from_override = not templates

    if from_override:
        template_path = global_config.defaults.path / global_config.override_config_path
        if not template_path.exists():
            typer.echo(
                "Template argument not specified and "
                "`.devcontainer/overrides.yaml` does not exist"
            )
            raise typer.Exit(1)
        templates = [template_path.as_posix()]
    else:
        templates = [global_config.load_alias_config().aliases.get(t, t) for t in templates]

    configs = [Config.parse_file(t).merge_bases() for t in templates]
    merged_config = global_config.merge_bases().defaults | reduce(lambda a, b: a | b, configs)

    COOKIECUTTER_CONFIG.write_text(
        merged_config.resolve().json(indent=4, exclude={"base_config": True})
    )
    cookiecutter(TEMPLATE_DIR.as_posix(), no_input=True, overwrite_if_exists=True)

    if not from_override:
        config_paths = [cfg.config_path.resolve() for cfg in configs]
        override_config = Config.NONE
        override_config.base_config = config_paths
        override_config.write_yaml(
            global_config.defaults.path / global_config.override_config_path
        )

    devcontainer_dir = Path(merged_config.path)
    if not merged_config.docker.file:
        (devcontainer_dir / "devcontainer.Dockerfile").unlink()
        (devcontainer_dir / "build.sh").unlink()

    if build:
        subprocess.run(["bash", ".devcontainer/build.sh"], check=True)


@app.command()
def create_template(
    template_path: str = typer.Argument(...),
    force: bool = typer.Option(False, "--force", "-f"),
    with_descriptions: bool = typer.Option(True, "-d/-D", "--with-description/--no-description"),
):
    global_config = GlobalConfig.load(create_if_not_exist=True)

    path = Path(template_path)

    global_template = False
    if path.suffix not in [".yaml", ".yml"]:
        global_template = True
        path = path.with_suffix(".yaml")

    if global_template:
        if global_config.template_dir.is_absolute():
            typer.echo("Error: --local not specified but path is absolute", err=True)
            raise typer.Exit(1)
        path = global_config.config_path.parent / global_config.template_dir / path

    if path.exists() and not force:
        if not typer.confirm(f"Config '{path}' already exists, overwrite?"):
            raise typer.Exit(1)

    Config().write_yaml(path, with_descriptions=with_descriptions)
    typer.echo("Config written to " f"'{typer.style(path, typer.colors.GREEN)}'")

    if global_template:
        alias_config = global_config.load_alias_config()
        alias_config.aliases[path.stem] = path.as_posix()
        alias_config.write_yaml()
        typer.echo("Alias created: " f"'{typer.style(path.stem, fg=typer.colors.BLUE)}'")


@app.command()
def version():
    typer.echo(__version__)


@app.callback(invoke_without_command=True)
def _main(version: bool = typer.Option(False, "--version", "-v")):
    if version:
        typer.echo(__version__)
        raise typer.Exit(0)


def main():
    app()


if __name__ == "__main__":
    main()



import re
from typing import Optional

DEVCONTAINER_MOUNT_STR_REGEX = re.compile(r"^src=(?P<src>.+),dst=(?P<dst>.+)(?P<args>,.+)?")
SIMPLE_MOUNT_REGEX = re.compile(r"(?P<src>.+):(?P<dst>.+)")


class MountString(str):
    __slots__ = ("src", "dst")

    def __new__(cls, v: Optional[str], **kwargs):
        return str.__new__(cls, v if v is not None else cls.build(**kwargs))

    def __init__(
        self,
        mount_string: str,
        *,
        src: str = None,
        dst: str = None,
    ):
        str.__init__(mount_string)
        if src is None or dst is None:
            arg_dict = type(self).validate_string(mount_string)
            src = arg_dict["src"]
            dst = arg_dict["dst"]

        self.src = src
        self.dst = dst

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: str):
        if value.__class__ == cls:
            return value

        if not isinstance(value, str):
            raise TypeError("mount must be string")

        return cls(value, **cls.validate_string(value))

    @classmethod
    def validate_string(cls, value: str):
        if value == "{{ base }}":
            return dict(src="{{ base }}", dst="{{ base }}")

        match = SIMPLE_MOUNT_REGEX.match(value)
        if match is None:
            match = DEVCONTAINER_MOUNT_STR_REGEX.match(value)
        if match is None:
            raise Exception()

        return dict(src=match.group("src"), dst=match.group("dst"))

    @classmethod
    def build(cls, *, src: str, dst: str):
        return f"{src}:{dst}"

    def to_devcontainer_format(self):
        return MountString(
            f"src={self.src}," f"dst={self.dst}," "type=bind,consistency=cached",
        )

    def __repr__(self) -> str:
        return f"MountString({self})"



import copy
import subprocess
from pathlib import Path

import jinja2


def render_recursive_template(template: str, values: dict):
    prev = template
    while True:
        rendered = jinja2.Template(prev, undefined=jinja2.StrictUndefined).render(values)
        if rendered == prev:
            return rendered
        prev = rendered


def get_project_root_basename() -> str:
    try:
        git_folder_path = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            check=True,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        ).stdout
        return Path(git_folder_path).name.strip()
    except subprocess.CalledProcessError:
        return None


def dict_merge(d, overwrite):
    new_config = copy.deepcopy(d)

    for k, v in overwrite.items():
        if isinstance(v, dict):
            orig_dict = d[k] if k in d and d[k] else dict()
            new_config[k] = dict_merge(orig_dict, v)
        elif isinstance(v, list):
            orig_list = d[k] if k in d and d[k] else list()
            new_config[k] = orig_list + v
        else:
            if v is not None:
                new_config[k] = v

    return new_config



class InvalidArgumentException(Exception):
    pass


class ConfigDoesNotExistException(Exception):
    pass


class ConfigMovedException(Exception):
    pass


class InvalidMountStringException(Exception):
    def __init__(self, mnt_str, *args: object) -> None:
        super().__init__(f"Mount string is invalid: '{mnt_str}'", *args)



from pathlib import Path

from pydantic import BaseSettings


class Settings(BaseSettings):
    global_config_filename = "config.yaml"
    global_config_dir = Path().home() / ".devcontainer_manager"

    @property
    def global_config_path(self):
        return self.global_config_dir / self.global_config_filename

    class Config:
        env_prefix = "devcontainer_manager_"
        case_insensitive = True



#!/usr/bin/env python
from pathlib import Path

from setuptools import find_packages, setup

from devcontainer_manager import __version__ as version

requirements = [req for req in Path("requirements.txt").read_text().splitlines() if req]
dev_requirements = [
    req
    for req in Path("requirements-dev.txt").read_text().splitlines()
    if req and not req.startswith("-")
]

setup(
    name="devcontainer_manager",
    version=version,
    author="gnox",
    url="https://github.com/gnox/devcontainer-manager",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=requirements,
    python_requires=">=3.6",
    extras_require={"dev": dev_requirements},
    license="MIT",
    entry_points={
        "console_scripts": ["devcontainer_manager = devcontainer_manager.__main__:main"]
    },
    packages=find_packages(),
    package_dir={"devcontainer_manager": "devcontainer_manager"},
    include_package_data=True,
)



from pathlib import Path
from typing import Dict, Optional

from pydantic import Field, validator

from .base_config import BaseYamlConfigModel, default_if_none

DEFAULT_ALIAS_FILENAME = "aliases.yaml"


class AliasConfig(BaseYamlConfigModel):
    aliases: Optional[Dict[str, Path]] = Field(
        default_factory=dict,
        description=(
            "aliases for templates, can be added using,\n" "`devcontainer_manager alias --help`"
        ),
    )

    _not_none = validator("*", pre=True, allow_reuse=True)(default_if_none)

    def resolve(self, alias):
        return self.aliases.get(alias, alias)
